#!/usr/bin/env python

import argparse
import cloudpickle

from io import open


def run(pickled_runner):
    f = cloudpickle.load(open(pickled_runner, "rb"))
    res = f["func"](*f["args"], **f["kwargs"])
    cloudpickle.dump(res, open("./result.pickle", "wb"))
    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("pickled_runner")
    args = parser.parse_args()
    run(args.pickled_runner)



from __future__ import absolute_import, print_function, unicode_literals

import cloudpickle
import copy
import os
import pkg_resources
import re
import sys
import tempfile
import tes
import uuid

from attr import attrs, attrib, Factory
from attr.validators import instance_of, optional
from builtins import str, bytes
from collections import Callable
from concurrent.futures import ThreadPoolExecutor
from io import open
from requests.utils import urlparse
from tes.models import strconv

from tesseract.filestore import FileStore
from tesseract.utils import process_url


@attrs
class Tesseract(object):
    file_store = attrib(validator=instance_of(FileStore))
    tes_url = attrib(default="http://localhost:8000", convert=strconv)
    __tes_client = attrib(init=False,  validator=instance_of(tes.HTTPClient))
    __id = attrib(init=False, convert=strconv, validator=instance_of(str))
    input_files = attrib(
        default=Factory(list), validator=tes.models.list_of(tes.TaskParameter)
    )
    output_files = attrib(
        default=Factory(list), validator=tes.models.list_of(tes.TaskParameter)
    )
    cpu_cores = attrib(default=None, validator=optional(instance_of(int)))
    ram_gb = attrib(
        default=None, validator=optional(instance_of((int, float)))
    )
    disk_gb = attrib(
        default=None, validator=optional(instance_of((int, float)))
    )
    docker = attrib(
        convert=strconv, validator=instance_of(str)
    )
    libraries = attrib(
        convert=strconv, validator=tes.models.list_of(str)
    )

    @docker.default
    def __default_docker(self):
        v = sys.version_info
        return "python:%s.%s.%s" % (v.major, v.minor, v.micro)

    @libraries.default
    def __default_libraries(self):
        return ["cloudpickle"]

    @tes_url.validator
    def __validate_tes_url(self, attribute, value):
        value = urlparse(value).geturl()
        if not isinstance(value, str):
            raise ValueError("%s must be a valid URL" % (attribute))

    def __attrs_post_init__(self):
        self.__tes_client = tes.HTTPClient(self.tes_url)
        self.__id = None

    def __get_id(self):
        if self.__id is None or self.file_store.exists(self.__id):
            self.__id = "tesseract_%s" % (uuid.uuid4().hex)
        return self.__id

    def clone(self):
        return copy.deepcopy(self)

    def with_resources(self, cpu_cores=None, ram_gb=None, disk_gb=None,
                       docker=None, libraries=[]):
        # only override if set
        self.cpu_cores = cpu_cores or self.cpu_cores
        self.ram_gb = ram_gb or self.ram_gb
        self.disk_gb = disk_gb or self.disk_gb
        self.docker = docker or self.docker
        self.libraries = libraries or self.libraries

    def with_input(self, url, path):
        u = urlparse(process_url(url))
        if u.scheme not in self.file_store.supported:
            raise ValueError(
                "Unsupported scheme - must be one of %s" %
                (self.file_store.supported)
            )
        if u.scheme == "file" and self.file_store.scheme != "file":
            raise ValueError("please upload your input file to the file store")

        u = urlparse(path)
        if u.scheme not in ["file", ""]:
            raise ValueError("runtime path must be defined as a local path")
        if u.scheme == "file":
            path = os.path.join(u.netloc, u.path)

        if path.startswith("./"):
            pass
        elif not os.path.isabs(path):
            raise ValueError(
                "runtime path must be an absolute path or start with './'"
            )

        self.input_files.append(
            tes.TaskParameter(
                path=os.path.join("/tmp/tesseract", re.sub("^./", "", path)),
                url=url,
                type="FILE"
            )
        )

    def with_output(self, path):
        u = urlparse(path)
        if u.scheme not in ["file", ""]:
            raise ValueError("runtime path must be defined as a local path")
        if u.scheme == "file":
            path = os.path.join(u.netloc, u.path)

        if path.startswith("./"):
            pass
        elif not os.path.isabs(path):
            raise ValueError(
                "runtime path must be an absolute path or start with './'"
            )

        run_id = self.__get_id()
        self.output_files.append(
            tes.TaskParameter(
                path=os.path.join("/tmp/tesseract", re.sub("^./", "", path)),
                url=self.file_store.generate_url(
                    os.path.join(run_id, re.sub("^./|^/", "", path))
                ),
                type="FILE"
            )
        )

    def run(self, func, *args, **kwargs):
        if not isinstance(func, Callable):
            raise TypeError("func not an instance of collections.Callable")

        run_id = self.__get_id()

        # serialize function and arguments
        # upload to object store if necessary
        runner = {"func": func, "args": args, "kwargs": kwargs}
        cp_str = cloudpickle.dumps(runner)
        input_name = os.path.join(run_id, "tesseract_func.pickle")
        input_cp_url = self.file_store.upload(name=input_name, contents=cp_str)

        # define storage url for pickled output
        output_cp_url = self.file_store.generate_url(
            "%s/tesseract_result.pickle" % (run_id)
        )

        # create task msg and submit
        task_msg = self._create_task_msg(input_cp_url, output_cp_url)
        id = self.__tes_client.create_task(task_msg)
        return Future(
            id,
            output_cp_url,
            self.file_store,
            self.__tes_client
        )

    def _create_task_msg(self, input_cp_url, output_cp_url):
        runner = pkg_resources.resource_string(
            __name__, "resources/runner.py"
        )
        if isinstance(runner, bytes):
            runner = runner.decode("utf8")

        cmd_install_reqs = "pip install %s" % (" ".join(self.libraries))
        cmd_tesseract = "python tesseract.py func.pickle"

        if len(self.libraries) == 0:
            cmd = cmd_tesseract
        else:
            cmd = cmd_install_reqs + " && " + cmd_tesseract

        task = tes.Task(
            name="tesseract remote execution",
            inputs=self.input_files + [
                tes.TaskParameter(
                    name="pickled function",
                    url=input_cp_url,
                    path="/tmp/tesseract/func.pickle",
                    type="FILE"
                ),
                tes.TaskParameter(
                    name="tesseract runner script",
                    path="/tmp/tesseract/tesseract.py",
                    type="FILE",
                    contents=str(runner)
                )
            ],
            outputs=self.output_files + [
                tes.TaskParameter(
                    name="pickled result",
                    url=output_cp_url,
                    path="/tmp/tesseract/result.pickle",
                    type="FILE"
                )
            ],
            resources=tes.Resources(
                cpu_cores=self.cpu_cores,
                ram_gb=self.ram_gb,
                size_gb=self.disk_gb
            ),
            executors=[
                tes.Executor(
                    image_name=self.docker,
                    cmd=["sh", "-c", cmd],
                    stdout="/tmp/tesseract/stdout",
                    stderr="/tmp/tesseract/stderr",
                    workdir="/tmp/tesseract"
                )
            ]
        )
        return task


@attrs
class Future(object):
    __id = attrib(convert=strconv, validator=instance_of(str))
    __output_url = attrib(convert=strconv, validator=instance_of(str))
    __file_store = attrib(validator=instance_of(FileStore))
    __client = attrib(validator=instance_of(tes.HTTPClient))
    __execption = attrib(
        init=False, default=None, validator=optional(instance_of(Exception))
    )
    __result = attrib(init=False, default=None)

    def __attrs_post_init__(self):
        pool = ThreadPoolExecutor(1)
        self.__result = pool.submit(self.__poll)

    def __poll(self):
        r = self.__client.wait(self.__id)
        if r.state != "COMPLETE":
            r = self.__client.get_task(self.__id, "FULL")
            raise RuntimeError("remote job failed:\n%s" % (r))

        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.close()
        self.__file_store.download(self.__output_url, tmp.name, True)
        return cloudpickle.load(open(tmp.name, "rb"))

    def result(self, timeout=None):
        return self.__result.result(timeout=timeout)

    def exeception(self, timeout=None):
        return self.__result.exception(timeout=timeout)

    def running(self):
        r = self.__client.get_task(self.__id, "MINIMAL")
        return r.state in["QUEUED", "INITIALIZING", "RUNNING"]

    def done(self):
        r = self.__client.get_task(self.__id, "MINIMAL")
        return r.state in ["COMPLETE", "ERROR", "SYSTEM_ERROR", "CANCELED"]

    def cancel(self):
        self.__result.cancel()
        self.__client.cancel(self.__id)
        return True

    def cancelled(self):
        r = self.__client.get_task(self.__id, "MINIMAL")
        return r.state == "CANCELED"



from __future__ import absolute_import, print_function, unicode_literals

import errno
import json
import os
import re

from libcloud.storage.types import Provider
from io import open
from tes.models import strconv
from requests.utils import urlparse


def makedirs(path, exists_ok=True):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            if exists_ok:
                pass
            else:
                raise exc
        else:
            raise exc


def process_url(value):
    value = strconv(value)
    u = urlparse(value)
    if u.scheme == "":
        url = "file://" + os.path.abspath(value)
    elif u.scheme == "file" and u.netloc != "":
        url = "file://" + os.path.abspath(os.path.join(u.netloc, u.path))
    else:
        url = value
    return url


def lookup_provider(scheme, region):
    lookup = {
        "gs": {
            "*": Provider.GOOGLE_STORAGE
        },
        "s3": {
            "us-east-1": Provider.S3,
            "us-east-2": Provider.S3_US_EAST2,
            "us-west-1": Provider.S3_US_WEST,
            "us-west-2": Provider.S3_US_WEST_OREGON
        },
        "swift": {
            "*": Provider.OPENSTACK_SWIFT
        }
    }

    if scheme in ["gs", "swift"]:
        region = "*"

    try:
        provider = lookup[scheme][region]
    except:
        raise RuntimeError(
            "%s provider for region %s could not be found" % (scheme, region)
        )
    return provider


def lookup_credentials(scheme):
    lookup = {
        "gs": "~/.config/gcloud/application_default_credentials.json",
        "s3": "~/.aws/credentials",
    }

    key = None
    secret = None

    try:
        if scheme == "swift":
            key = os.environ["OS_USERNAME"]
            secret = os.environ["OS_PASSWORD"]
            if key is None or secret is None:
                raise

        elif scheme in lookup:
            fh = open(os.path.expanduser(lookup[scheme]), "r")
            content = fh.read()
            fh.close()
            if scheme == "gs":
                cred = json.loads(content)
                key = cred["client_id"]
                secret = key = cred["client_secret"]

            elif scheme == "s3":
                key = re.findall(
                    '(aws_access_key_id\ =\ )(.*)\n?', content
                )[0][1]
                secret = re.findall(
                    '(aws_secret_access_key\ =\ )(.*)\n?', content
                )[0][1]
    except:
        raise RuntimeError(
            "%s credentials could not be set automatically, " % (scheme) +
            "please provide your key and secret"
        )

    return key, secret


def lookup_region(scheme):
    lookup = {
        "s3": "~/.aws/config",
    }

    region = None
    # return none if scheme isn't 's3'
    if scheme not in lookup:
        return region

    try:
        fh = open(os.path.expanduser(lookup[scheme]), "r")
        if scheme == "s3":
            content = fh.read()
            region = re.findall(
                '(region\ =\ )(.*)\n?', content
            )[0][1]
        fh.close()
    except:
        raise RuntimeError(
            "%s region could not be set automatically, " % (scheme) +
            "please provide the region to use for your bucket"
        )
    return region


def lookup_project(scheme):
    lookup = {
        "gs": "~/.config/gcloud/configurations/config_default",
    }

    project = None
    # return none if scheme isn't 'gs'
    if scheme not in lookup:
        return project

    try:
        fh = open(os.path.expanduser(lookup[scheme]), "r")
        if scheme == "gs":
            content = fh.read()
            project = re.findall(
                '(project\ =\ )(.*)\n?', content
            )[0][1]
        fh.close()
    except:
        raise RuntimeError(
            "%s project could not be set automatically, " % (scheme) +
            "please provide the project to use for your bucket"
        )
    return project



from __future__ import absolute_import, print_function, unicode_literals


import os
import re
import shutil
import six
import tempfile
import uuid

from attr import attrs, attrib
from attr.validators import instance_of, optional
from builtins import str, bytes
from libcloud.storage.providers import get_driver
from io import open
from tes.models import strconv
from requests.utils import urlparse

from tesseract.utils import (makedirs, process_url, lookup_provider,
                             lookup_credentials, lookup_region,
                             lookup_project)


@attrs
class FileStore(object):
    filestore_url = attrib(convert=process_url)
    key = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    secret = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    region = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    project = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    ex_force_auth_url = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    ex_force_auth_version = attrib(
        default='2.0_password',
        convert=strconv,
        validator=optional(instance_of(str))
    )
    ex_tenant_name = attrib(
        default=None, convert=strconv, validator=optional(instance_of(str))
    )
    provider = attrib(init=False)
    driver = attrib(init=False)
    scheme = attrib(init=False, convert=strconv, validator=instance_of(str))
    __bucket = attrib(init=False, convert=strconv, validator=instance_of(str))
    __path = attrib(init=False, convert=strconv, validator=instance_of(str))
    supported = ["file", "gs", "s3", "swift"]

    @filestore_url.validator
    def __validate_filestore_url(self, attribute, value):
        u = urlparse(value)
        if u.scheme == "file" and u.netloc != "":
            raise ValueError(
                "invalid url"
            )
        if u.scheme not in self.supported:
            raise ValueError(
                "Unsupported scheme - must be one of %s" % (self.supported)
            )

    def __attrs_post_init__(self):
        u = urlparse(self.filestore_url)
        self.scheme = u.scheme
        self.__bucket = u.netloc
        self.__path = re.sub("/$", "", u.path)

        if self.scheme != "file":
            self.provider = get_driver(
                lookup_provider(self.scheme, self.region)
            )

            if self.key is None and self.secret is None:
                self.key, self.secret = lookup_credentials(self.scheme)

            # Openstack Swift
            if self.scheme == "swift":
                try:
                    auth_url = urlparse(os.environ['OS_AUTH_URL'])
                    self.ex_force_auth_url = "%s://%s" % (auth_url.scheme,
                                                          auth_url.netloc)
                    self.ex_tenant_name = os.environ['OS_TENANT_NAME']
                except:
                    raise ValueError(
                        "OS_AUTH_URL and/or OS_TENANT_NAME were not found"
                    )
                self.driver = self.provider(
                    key=self.key,
                    secret=self.secret,
                    ex_force_auth_url=self.ex_force_auth_url,
                    ex_tenant_name=self.ex_tenant_name,
                    ex_force_auth_version=self.ex_force_auth_version,
                )

            # Google Storage or Amazon S3
            else:
                if self.region is None:
                    self.region = lookup_region(self.scheme)

                if self.project is None:
                    self.project = lookup_project(self.scheme)

                self.driver = self.provider(
                    key=self.key,
                    secret=self.secret,
                    region=self.region,
                    project=self.project
                )

        self.key = None
        self.secret = None
        self.__create_store()

    def __create_store(self):
        if self.scheme == "file":
            makedirs(self.__path, exists_ok=True)
        else:
            try:
                self.driver.get_container(self.__bucket)
            except:
                self.driver.create_container(self.__bucket)
        return

    def __delete_bucket(self):
        if self.scheme == "file":
            shutil.rmtree(self.__bucket)
        else:
            self.driver.delete_container(self.__bucket)
        return

    def generate_url(self, name):
        return "%s://%s" % (
            self.scheme, os.path.join(self.__bucket, self.__path, name)
        )

    def exists(self, name):
        if os.path.basename(name) != name:
            raise ValueError("must pass only the name of a file or directory")
        found = False

        if self.scheme == "file":
            for root, dirnames, filenames in os.walk(self.__path):
                if name in dirnames + filenames + [os.path.basename(root)]:
                    found = True
        else:
            objs = self.driver.list_container_objects(
                self.driver.get_container(self.__bucket)
            )
            for o in objs:
                if name in o.name:
                    found = True

        return found

    def upload(self, path=None, name=None, contents=None,
               overwrite_existing=False):
        if path is not None and contents is not None:
            raise ValueError("Cannot provide both local path and contents")
        if path is None and contents is None:
            raise ValueError("Provide either local path or contents")

        if name is None:
            if path is None:
                name = os.path.basname(path)
            else:
                name = str("tmp_" + uuid.uuid4().hex)

        if path is not None:
            contents = open(path, "r").read()

        if isinstance(contents, six.string_types):
            contents = bytes(contents, "utf8")

        url = os.path.join(self.__path, name)

        if self.scheme == "file":
            if os.path.exists(url) and not overwrite_existing:
                raise OSError(
                    "File exists; to force set overwrite_existing to True"
                )
            makedirs(os.path.dirname(url), exists_ok=True)
            with open(url, "wb") as fh:
                fh.write(contents)
        else:
            url = re.sub("^/", "", url)
            tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)
            tmpf.write(contents)
            tmpf.close()
            self.driver.upload_object(
                file_path=tmpf.name,
                container=self.driver.get_container(self.__bucket),
                object_name=url
            )
            os.remove(tmpf.name)

        return url

    def download(self, url, destination_path, overwrite_existing=False):
        if os.path.exists(destination_path) and not overwrite_existing:
            raise OSError(
                "destination_path exists, and overwrite_existing = False"
            )
        u = urlparse(url)
        if u.scheme not in self.supported:
            raise ValueError(
                "Unsupported scheme - must be one of %s" % (self.supported)
            )
        source = u.path
        if u.scheme == "file":
            shutil.copyfile(source, destination_path)
        else:
            source = re.sub("^/", "", source)
            obj = self.driver.get_object(self.__bucket, source)
            self.driver.download_object(
                obj,
                destination_path,
                overwrite_existing
            )
        return destination_path



#!/usr/bin/env python

import io
import os
import re

from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='py-tesseract',
    version=find_version("tesseract", "__init__.py"),
    description='Remote code execution with the GA4GH Task Execution API',
    author='OHSU Computational Biology',
    author_email='CompBio@ohsu.edu',
    maintainer='Adam Struck',
    maintainer_email='strucka@ohsu.edu',
    url="https://github.com/ohsu-comp-bio/tesseract",
    license='MIT',
    packages=["tesseract"],
    package_data={"tesseract": ["resources/*.py"]},
    python_requires='>=2.6, <4',
    install_requires=[
        "apache_libcloud>=2.1.0",
        "attrs>=17.2.0",
        "cloudpickle>=0.3.1",
        "future>=0.16.0",
        "futures>=3.1.1",
        "py-tes>=0.1.5",
        "requests>=2.18.1"
    ],
    zip_safe=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from __future__ import absolute_import, print_function

from tesseract.tesseract import Tesseract
from tesseract.filestore import FileStore

__all__ = [
    Tesseract,
    FileStore
]

__version__ = "0.1.0"
